using System.Collections;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;

namespace SpiderWeb.Controls
{
	[TemplatePart(Name="PanningContainer", Type=typeof(Canvas))]
	[TemplatePart(Name="ItemsContainer", Type=typeof(Canvas))]
	[TemplatePart(Name="ConnectorsContainer", Type=typeof(Canvas))]
	public class SpiderWebControl : Control
	{
		public static readonly DependencyProperty ConnectorStyleProperty = DependencyProperty.Register(
											"ConnectorStyle", typeof(Style), typeof(SpiderWebControl), null);

		public static readonly DependencyProperty HeaderProperty = DependencyProperty.Register(
			"Header", typeof(object), typeof(SpiderWebControl), new PropertyMetadata(OnHeaderChanged));

		public static readonly DependencyProperty ItemsSourceProperty = DependencyProperty.Register(
			"ItemsSource", typeof(IEnumerable), typeof(SpiderWebControl), null);

		public static readonly DependencyProperty ItemTemplateProperty = DependencyProperty.Register(
			"ItemTemplate", typeof(DataTemplate), typeof(SpiderWebControl), null);

		public static readonly DependencyProperty SelectedItemProperty = DependencyProperty.Register(
			"SelectedItem", typeof(object), typeof(SpiderWebControl), null);

		private SpiderWebItem _selectedContainer;
		private bool _isPanning;
		private Point _startPosition;

		public object SelectedItem
		{
			get { return GetValue(SelectedItemProperty); }
			set { SetValue(SelectedItemProperty, value); }
		}

		public DataTemplate ItemTemplate
		{
			get { return (DataTemplate)GetValue(ItemTemplateProperty); }
			set { SetValue(ItemTemplateProperty, value); }
		}

		protected SpiderWebItem RootItem { get; set; }

		public object Header
		{
			get { return GetValue(HeaderProperty); }
			set { SetValue(HeaderProperty, value); }
		}

		protected Canvas ConnectorsContainer { get; set; }

		protected Canvas ItemsContainer { get; set; }

		public Style ConnectorStyle
		{
			get { return (Style)GetValue(ConnectorStyleProperty); }
			set { SetValue(ConnectorStyleProperty, value); }
		}

		public IEnumerable ItemsSource
		{
			get { return (IEnumerable)GetValue(ItemsSourceProperty); }
			set { SetValue(ItemsSourceProperty, value); }
		}

		protected Canvas PanningContainer { get; set; }

		public SpiderWebControl()
		{
			DefaultStyleKey = typeof(SpiderWebControl);
		}


		private static void OnHeaderChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
		{
			var control = d as SpiderWebControl;
			var header = e.NewValue;

			control.SetRootItem(header);
		}

		private void SetRootItem(object newValue)
		{
			if (RootItem != null)
			{
				RootItem.Owner = null;
				RootItem.ClearValue(SpiderWebItem.ItemsSourceProperty);
				ItemsContainer.Children.Remove(RootItem);
			}

			if (ItemsContainer != null)
			{
				RootItem = new SpiderWebItem()
				{
					Content = newValue,
					ContentTemplate = ItemTemplate,
					Owner = this
				};

				if (ItemTemplate is HierarchicalDataTemplate)
				{
					var binding = (ItemTemplate as HierarchicalDataTemplate).ItemsSource;
					BindingOperations.SetBinding(RootItem, SpiderWebItem.ItemsSourceProperty, binding);
				}

				ItemsContainer.Children.Add(RootItem);
			}
		}

		private void OnPanStart(object sender, MouseButtonEventArgs e)
		{
			if ((Keyboard.Modifiers & ModifierKeys.Control) == ModifierKeys.Control)
			{
				_isPanning = true;
				_startPosition = e.GetPosition(this);
				PanningContainer.CaptureMouse();
			}
		}

		private void OnPanEnd(object sender, MouseButtonEventArgs e)
		{
			_isPanning = false;
			PanningContainer.ReleaseMouseCapture();
		}

		private void OnPanMove(object sender, MouseEventArgs e)
		{
			var isControlPressed = (Keyboard.Modifiers & ModifierKeys.Control) == ModifierKeys.Control;
			if (_isPanning && isControlPressed)
			{
				var currentPosition = e.GetPosition(this);
				var delta = new Point(currentPosition.X - _startPosition.X, currentPosition.Y - _startPosition.Y);
				_startPosition = currentPosition;

				var connXform = ConnectorsContainer.RenderTransform as TranslateTransform;
				var itemXform = ItemsContainer.RenderTransform as TranslateTransform;

				connXform.X += delta.X;
				connXform.Y += delta.Y;

				itemXform.X += delta.X;
				itemXform.Y += delta.Y;
			}
		}

		public override void OnApplyTemplate()
		{
			PanningContainer = GetTemplateChild("PanningContainer") as Canvas;
			PanningContainer.MouseLeftButtonDown += OnPanStart;
			PanningContainer.MouseMove += OnPanMove;
			PanningContainer.MouseLeftButtonUp += OnPanEnd;

			ConnectorsContainer = GetTemplateChild("ConnectorsContainer") as Canvas;
			ConnectorsContainer.RenderTransform = new TranslateTransform();

			ItemsContainer = GetTemplateChild("ItemsContainer") as Canvas;
			ItemsContainer.RenderTransform = new TranslateTransform();

			SetRootItem(Header);
		}

		internal void AddItemContainer(SpiderWebItem container, Line connector)
		{
			ItemsContainer.Children.Add(container);
			ConnectorsContainer.Children.Add(connector);
		}

		public void RemoveItemContainer(SpiderWebItem container, Line connector)
		{
			ItemsContainer.Children.Remove(container);
			ConnectorsContainer.Children.Remove(connector);
		}

		public void SetSelection(SpiderWebItem spiderWebItem)
		{
			if (_selectedContainer == spiderWebItem) return; // clicked on already selected instance

			if (_selectedContainer != null)
			{
				_selectedContainer.IsSelected = false;
			}

			_selectedContainer = spiderWebItem;
			_selectedContainer.IsSelected = true;
			SelectedItem = spiderWebItem.Content;
		}

		public void ApplyLayout(ILayout layout)
		{
			if (layout != null)
			{
				layout.Apply(new Rect(new Point(), new Size(ActualWidth, ActualHeight)), RootItem);
				foreach (var pair in layout.LayoutPositions)
				{
					var item = pair.Key;
					var position = pair.Value;

					item.AnimationPosition(position);
				}
			}
		}
	}
}