﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace SpiderWeb.Controls
{
	[TemplateVisualState(GroupName = "HoverStates", Name = "NoHover")]
	[TemplateVisualState(GroupName = "HoverStates", Name = "Hover")]
	[TemplateVisualState(GroupName = "SelectionStates", Name = "Unselected")]
	[TemplateVisualState(GroupName = "SelectionStates", Name = "Selected")]
	public class SpiderWebItem : ContentControl
	{
		private bool _isDragging;
		private Point _startPosition;

		private Storyboard _locationAnimator;
		private PointAnimation _pointAnimation;
		private bool _isAnimating;

		public static readonly DependencyProperty ItemsSourceProperty = DependencyProperty.Register(
			"ItemsSource", typeof(IEnumerable), typeof(SpiderWebItem), new PropertyMetadata(OnItemsSourceChanged));

		public static readonly DependencyProperty IsSelectedProperty = DependencyProperty.Register(
			"IsSelected", typeof (bool), typeof (SpiderWebItem), new PropertyMetadata(OnIsSelectedChanged));

		public static readonly DependencyProperty LocationProperty = DependencyProperty.Register(
			"Location", typeof (Point), typeof (SpiderWebItem), new PropertyMetadata(OnLocationChanged));

		private static readonly DependencyProperty LinkedConnectorProperty = DependencyProperty.Register(
			"LinkedConnector", typeof(Line), typeof(SpiderWebItem), null);

		private Dictionary<object, SpiderWebItem> ItemMap { get; set; }
		internal bool HasChildren
		{
			get
			{
				return ItemMap != null && ItemMap.Values.Count > 0;
			}
		}
		internal List<SpiderWebItem> ChildItems
		{
			get
			{
				return ItemMap.Values.ToList();
			}
		}

		public bool IsSelected
		{
			get { return (bool) GetValue(IsSelectedProperty); }
			set { SetValue(IsSelectedProperty, value); }
		}

		public IEnumerable ItemsSource
		{
			get { return (IEnumerable)GetValue(ItemsSourceProperty); }
			set { SetValue(ItemsSourceProperty, value); }
		}

		public Point Location
		{
			get { return (Point) GetValue(LocationProperty); }
			set { SetValue(LocationProperty, value); }
		}

		internal SpiderWebItem ParentItem { get; set; }
		internal SpiderWebControl Owner { get; set; }

		internal bool IsConnectorInitialized { get; set; }

		public bool IsMouseOver { get; set; }


		public SpiderWebItem()
		{
			DefaultStyleKey = typeof(SpiderWebItem);

			InitializeStoryboard();
		}

		private void InitializeStoryboard()
		{
			_pointAnimation = new PointAnimation()
			                  	{
			                  		To = new Point(),
			                  		Duration = new Duration(TimeSpan.FromSeconds(0.5)),
			                  		EasingFunction = new CubicEase()
			                  	};
			Storyboard.SetTarget(_pointAnimation, this);
			Storyboard.SetTargetProperty(_pointAnimation, new PropertyPath("Location"));

			_locationAnimator = new Storyboard();
			_locationAnimator.Children.Add(_pointAnimation);
			_locationAnimator.Completed += delegate
			                         	{
			                         		_isAnimating = false;
			                         		var pt = _pointAnimation.To.Value;
			                         		Location = pt;
			                         	};
		}

		private static void OnLocationChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
		{
			var item = d as SpiderWebItem;
			item.MoveItemTo((Point)e.NewValue);
		}

		private static void OnItemsSourceChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
		{
			var control = d as SpiderWebItem;
			var newSource = e.NewValue as IEnumerable;
			var oldSource = e.OldValue as IEnumerable;
			control.PopulateItems(oldSource, newSource);
		}

		private static void OnIsSelectedChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
		{
			var control = d as SpiderWebItem;
			control.UpdateVisualStates();
		}

		private void PopulateItems(IEnumerable oldSource, IEnumerable newSource)
		{
			if (oldSource != null)
			{
				ItemMap.Clear();
				ItemMap = null;

				if (oldSource is INotifyCollectionChanged)
				{
					var incc = oldSource as INotifyCollectionChanged;
					incc.CollectionChanged -= OnItemsCollectionChanged;
				}
			}

			if (newSource != null)
			{
				if (newSource is INotifyCollectionChanged)
				{
					var incc = newSource as INotifyCollectionChanged;
					incc.CollectionChanged += OnItemsCollectionChanged;
				}
				ItemMap = new Dictionary<object, SpiderWebItem>();
				foreach (var modelItem in newSource)
				{
					PrepareContainer(modelItem);
				}
			}
		}

		private void OnItemsCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			switch (e.Action)
			{
				case NotifyCollectionChangedAction.Add:
					foreach (var newItem in e.NewItems)
					{
						PrepareContainer(newItem);
					}
					break;
				case NotifyCollectionChangedAction.Remove:
					foreach (var oldItem in e.OldItems)
					{
						var container = ItemMap[oldItem];
						ClearContainer(container);
					}
					break;
			}
		}

		private void ClearContainer(SpiderWebItem container)
		{
			var connector = container.GetValue(LinkedConnectorProperty) as Line;
			Owner.RemoveItemContainer(container, connector);

			ItemMap.Remove(container.Content);

			container.Content = null;
			container.Owner = null;
			container.ParentItem = null;

			if (container.HasChildren)
			{
				foreach (var childContainer in container.ChildItems)
				{
					ClearContainer(childContainer);
				}
			}
		}

		private void PrepareContainer(object modelItem)
		{
			var container = new SpiderWebItem
								{
									Content = modelItem,
									Owner = Owner,
									ParentItem = this
								};

			// Template
			if (ContentTemplate is HierarchicalDataTemplate)
			{
				var hTemplate = ContentTemplate as HierarchicalDataTemplate;
				container.ContentTemplate = hTemplate.ItemTemplate;

				var binding = hTemplate.ItemsSource;
				BindingOperations.SetBinding(container, ItemsSourceProperty, binding);

				container.DataContext = modelItem;
			}

			// Connector
			var connector = new Line { Style = Owner.ConnectorStyle };
			container.SetValue(LinkedConnectorProperty, connector);

			// Item map
			ItemMap[modelItem] = container;
			Owner.AddItemContainer(container, connector);

			// Positioning
			Dispatcher.BeginInvoke(() =>
			                       	{
			                       		var location = new Point(Canvas.GetLeft(this), Canvas.GetTop(this));
			                       		location.X += ActualWidth;
			                       		location.Y += ActualHeight;
			                       		container.MoveItemBy(location, true);
			                       	});
		}

		protected override void OnMouseEnter(MouseEventArgs e)
		{
			IsMouseOver = true;
			UpdateVisualStates();
		}

		protected override void OnMouseLeave(MouseEventArgs e)
		{
			IsMouseOver = false;
			UpdateVisualStates();
		}

		protected override void OnMouseMove(MouseEventArgs e)
		{
			if (_isDragging)
			{
				var currentPosition = e.GetPosition(ParentItem);
				var delta = new Point(currentPosition.X - _startPosition.X, currentPosition.Y - _startPosition.Y);
				_startPosition = currentPosition;

				var moveGroup = (Keyboard.Modifiers & ModifierKeys.Control) == ModifierKeys.Control;
				MoveItemBy(delta, moveGroup);
			}
		}

		protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
		{
			_isDragging = true;

			CaptureMouse();
			_startPosition = e.GetPosition(ParentItem);
			Owner.SetSelection(this);

			UpdateVisualStates();
		}

		protected override void OnMouseLeftButtonUp(MouseButtonEventArgs e)
		{
			_isDragging = false;
			ReleaseMouseCapture();
		}

		private void UpdateVisualStates()
		{
			if (IsMouseOver)
			{
				VisualStateManager.GoToState(this, "Hover", true);
			}
			else
			{
				VisualStateManager.GoToState(this, "NoHover", true);
			}

			if (IsSelected)
			{
				VisualStateManager.GoToState(this, "Selected", true);
			}
			else
			{
				VisualStateManager.GoToState(this, "Unselected", true);
			}
		}

		private void MoveItemBy(Point delta, bool moveGroup)
		{
			var x = Canvas.GetLeft(this) + delta.X;
			var y = Canvas.GetTop(this) + delta.Y;

			Canvas.SetLeft(this, x);
			Canvas.SetTop(this, y);
			if (!_isAnimating)
			{
				Location = new Point(x, y);
			}

			UpdateConnector();

			if (moveGroup && ItemMap != null)
			{
				foreach (var childItem in ItemMap.Values)
				{
					childItem.MoveItemBy(delta, moveGroup);
				}
			}
		}

		private void MoveItemTo(Point point)
		{
			var currX = Canvas.GetLeft(this);
			var currY = Canvas.GetTop(this);
			var delta = new Point(point.X - currX, point.Y - currY);

			MoveItemBy(delta, false);
		}

		internal void AnimationPosition(Point point)
		{
			if (_isAnimating)
			{
				_locationAnimator.Pause();
			}

			_isAnimating = true;
			_pointAnimation.To = point;
			_locationAnimator.Begin();
		}

		private void UpdateConnector()
		{
			var connector = GetValue(LinkedConnectorProperty) as Line;
			if (ParentItem != null)
			{
				connector.X1 = Canvas.GetLeft(this) + ActualWidth / 2;
				connector.Y1 = Canvas.GetTop(this) + ActualHeight / 2;
				connector.X2 = Canvas.GetLeft(ParentItem) + ParentItem.ActualWidth / 2;
				connector.Y2 = Canvas.GetTop(ParentItem) + ParentItem.ActualHeight / 2;
			}
			if (ItemMap != null)
			{
				foreach (var connectedItem in ItemMap.Values)
				{
					var childConnector = connectedItem.GetValue(LinkedConnectorProperty) as Line;

					childConnector.X1 = Canvas.GetLeft(connectedItem) + connectedItem.ActualWidth / 2;
					childConnector.Y1 = Canvas.GetTop(connectedItem) + connectedItem.ActualHeight / 2;
					childConnector.X2 = Canvas.GetLeft(this) + ActualWidth / 2;
					childConnector.Y2 = Canvas.GetTop(this) + ActualHeight / 2;
				}
			}
		}
	}
}